"""Data Logger Callback."""

from typing import Optional, List

import numpy as np
import pandas as pd

from moving_targets.callbacks.callback import Callback
from moving_targets.util.typing import Dataset


class DataLogger(Callback):
    """Callback which stores the input data along with the predictions and adjustments after each iteration."""

    def __init__(self, filepath: Optional[str] = None):
        """
        :param filepath:
            Path string in which to write the stored data. Depending on the path extension, the data will be stored as
             a 'csv', 'tsv' or 'text' file. If None is passed, does not write any data.
        """
        super(DataLogger, self).__init__()

        self.filepath: Optional[str] = filepath
        """Path string in which to write the stored data."""

        self.data: pd.DataFrame = pd.DataFrame()
        """Dataframe object in which to store the input and output data."""

        self.iterations: List[int] = []
        """The list of macs iterations."""

    def on_process_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self.data = pd.concat((self.data, x), axis=1)
        self.data['y'] = y

    def on_iteration_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self.iterations.append(macs.iteration)

    def on_training_end(self, macs, x, y: np.ndarray, p: np.ndarray, val_data: Optional[Dataset]):
        self.data[f'p{macs.iteration}'] = p

    def on_adjustment_end(self, macs, x, y: np.ndarray, z: np.ndarray, val_data: Optional[Dataset]):
        self.data[f'z{macs.iteration}'] = z

    def on_process_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        if self.filepath is None:
            return
        elif self.filepath.endswith('.csv'):
            self.data.to_csv(self.filepath, index=False)
        elif self.filepath.endswith('.tsv'):
            self.data.to_csv(self.filepath, index=False, sep='\t')
        else:
            with open(self.filepath, 'w') as f:
                f.write(self.data.to_string())



"""Utility functions and classes for errors handling."""
import logging
from typing import Dict


def not_implemented_message(name: str, abstract: bool = True, static: bool = False) -> str:
    """Default error message for not implemented methods.

    :param name:
        The method name.

    :param abstract:
        Whether or not the method is abstract.

    :param static:
        Whether or not the method is static.

    :return:
        A string representing the error message.
    """

    abstract = 'abstract ' if abstract else ''
    static = 'static ' if static else ''
    return f"Please implement {static}{abstract}method '{name}'"


class MissingDependencyError(ModuleNotFoundError):
    """Error class for missing external dependencies which mat be lazily imported."""

    _versions: Dict[str, str] = {
        'cvxpy': '1.1.13',
        'docplex': '2.20.204',
        'gurobipy': '9.1.2',
        'tensorflow': '2.7.0',
        'wandb': '0.12.6'
    }
    """Expected versions (as in project requirements) of each external dependency which may be lazily imported."""

    def __init__(self, package: str):
        """
        :param package:
            The package name.
        """
        version = self._versions.get(package)
        if version is None:
            command = package
            logging.getLogger('MissingDependencyError').warning(f"unexpected lazy dependency '{package}'")
        else:
            command = f'{package}~={version}'
        message = f"This class requires '{package}' in order to be used, please install it via 'pip install {command}'"
        super(MissingDependencyError, self).__init__(message)


class BackendError(Exception):
    """Custom exception for unsupported `Backend` operations."""

    def __init__(self, unsupported: str, message: str = 'This backend cannot deal with'):
        """
        :param unsupported:
            The unsupported operation description.

        :param message:
            The prepended message.
        """
        super(BackendError, self).__init__(f'{message} {unsupported}')



from typing import Any, Union, List, Optional

import numpy as np

from moving_targets.masters.backends import Backend
from moving_targets.util.errors import MissingDependencyError, BackendError


class CplexBackend(Backend):
    """`Backend` implementation for the Cplex Solver."""

    def __init__(self, time_limit: Optional[float] = None):
        """
        :param time_limit:
            The solver time limit.
        """
        super(CplexBackend, self).__init__(sum_fn=lambda v: self.model.sum(v))

        try:
            import docplex.mp.model
            self._cp = docplex.mp.model
            """The lazily imported docplex instance."""
        except ModuleNotFoundError:
            raise MissingDependencyError(package='docplex')

        self.time_limit: Optional[float] = time_limit
        """The solver time limit."""

    def _build_model(self) -> Any:
        model = self._cp.Model(name='model')
        if self.time_limit is not None:
            model.set_time_limit(self.time_limit)
        return model

    def _solve_model(self) -> Optional:
        solution = self.model.solve()
        return solution

    def clear(self) -> Any:
        self.model.end()
        super(CplexBackend, self).clear()

    def maximize(self, cost) -> Any:
        self.model.maximize(cost)
        return self

    def minimize(self, cost) -> Any:
        self.model.minimize(cost)
        return self

    def get_objective(self) -> float:
        return self.solution.objective_value

    def get_values(self, expressions: np.ndarray) -> np.ndarray:
        return np.reshape([v.solution_value for v in expressions.flatten()], expressions.shape)

    def add_variable(self, vtype: str, lb: float, ub: float, name: Optional[str] = None) -> Any:
        # handle variable type and bounds
        if vtype == 'binary':
            assert lb == 0 and ub == 1, f"Binary variable type accepts [0, 1] bounds only, but [{lb}, {ub}] was passed."
            return self.model.binary_var(name=name)
        elif vtype == 'integer':
            return self.model.integer_var(lb=lb, ub=ub, name=name)
        elif vtype == 'continuous':
            return self.model.continuous_var(lb=lb, ub=ub, name=name)
        else:
            raise BackendError(unsupported=f"vtype '{vtype}'")

    def add_variables(self, *keys: int, vtype: str, lb: float, ub: float, name: Optional[str] = None) -> np.ndarray:
        if len(keys) == 0:
            # if no keys are passed, builds a single variable then reshape it into a zero-dimensional numpy array
            var = self.add_variable(vtype=vtype, lb=lb, ub=ub, name=name)
        elif vtype not in ['binary', 'integer', 'continuous']:
            # check correctness of type variables
            raise BackendError(unsupported=f"vtype '{vtype}'")
        else:
            # handle non-zero dimensionality
            if len(keys) == 1:
                fn = getattr(self.model, f'{vtype}_var_dict')
                kw = dict(keys=keys[0])
            elif len(keys) == 2:
                fn = getattr(self.model, f'{vtype}_var_matrix')
                kw = dict(keys1=keys[0], keys2=keys[1])
            elif len(keys) == 3:
                fn = getattr(self.model, f'{vtype}_var_cube')
                kw = dict(keys1=keys[0], keys2=keys[1], keys3=keys[2])
            else:
                raise BackendError(unsupported='variables having more than three dimensions')
            # and check the correctness of bounds wrt variable type
            if vtype == 'binary':
                assert lb == 0 and ub == 1, f"Binary variables accept [0, 1] bounds only, but [{lb}, {ub}] was passed."
            else:
                kw.update({'lb': lb, 'ub': ub})
            # handle variables
            var = fn(**kw, name=name)
            var = list(var.values())
        return np.array(var).reshape(keys)

    def add_constraints(self, constraints: Union[List, np.ndarray], name: Optional[str] = None) -> Any:
        self.model.add_constraints(constraints, names=name)
        return self

    def add_constraint(self, constraint, name: Optional[str] = None) -> Any:
        self.model.add_constraint(constraint, ctname=name)
        return self

    def add_indicator_constraints(self,
                                  indicators: np.ndarray,
                                  expressions: Union[List, np.ndarray],
                                  value: int = 1,
                                  name: Optional[str] = None) -> Any:
        binary_vars, cst = indicators.flatten(), np.array(expressions).flatten()
        self.model.add_indicators(binary_vars=binary_vars, cts=cst, true_values=value, names=name)
        return self

    def add_indicator_constraint(self,
                                 indicator: Any,
                                 expression: Any,
                                 value: int = 1,
                                 name: Optional[str] = None) -> Any:
        self.model.add_indicator(binary_var=indicator, linear_ct=expression, active_value=value, name=name)

    def square(self, a: np.ndarray, aux: Optional[str] = 'auto') -> np.ndarray:
        self._aux_warning(exp=None, aux=aux, msg='cannot impose equality constraints on quadratic expressions')
        return a ** 2

    def abs(self, a: np.ndarray, aux: Optional[str] = 'auto') -> np.ndarray:
        expressions = np.reshape([self.model.abs(v) for v in a.flatten()], a.shape)
        return self.aux(expressions=expressions, aux_vtype=aux)

    def divide(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = 'auto'):
        try:
            return super(CplexBackend, self).divide(a, b, aux=aux)
        except self._cp.DOcplexException:
            # docplex.mp.utils.DOcplexException: x1 / x11 : operation not supported, only numbers can be denominators
            # in case the divisor is not an array of constants, we cannot handle the case by adding the auxiliary
            # variables as in the gurobi backend because cplex cannot handle quadratic constraints
            raise BackendError(unsupported='divisions having variables in the denominator')



"""Moving Targets Backends for Masters."""

from moving_targets.masters.backends.backend import Backend
from moving_targets.masters.backends.cplex_backend import CplexBackend
from moving_targets.masters.backends.cvxpy_backend import CvxpyBackend
from moving_targets.masters.backends.gurobi_backend import GurobiBackend
from moving_targets.masters.backends.numpy_backend import NumpyBackend

aliases: dict = {
    'cvxpy': CvxpyBackend,
    'cplex': CplexBackend,
    'gurobi': GurobiBackend
}



"""Regression Metrics."""

from typing import Callable, Dict, Any

import numpy as np

from moving_targets.metrics.metric import Metric


class RegressionMetric(Metric):
    """Basic interface for a Moving Targets Regression Metric."""

    def __init__(self, metric_function: Callable, name: str, **metric_kwargs):
        """
        :param metric_function:
            Callable function that computes the metric given the true targets and the predictions.

        :param name:
            The name of the metric.

        :param metric_kwargs:
            Custom kwargs to be passed to the metric function.
        """
        super(RegressionMetric, self).__init__(name=name)

        self.metric_function: Callable = metric_function
        """The callable function used to compute the metric."""

        self.metric_kwargs: Dict[str, Any] = metric_kwargs
        """Custom arguments to be passed to the metric function."""

    def __call__(self, x, y: np.ndarray, p: np.ndarray) -> float:
        return self.metric_function(y, p, **self.metric_kwargs)


class MSE(RegressionMetric):
    """Wrapper for scikit-learn 'mean_squared_error' function."""

    def __init__(self, name: str = 'mse', **scikit_kwargs):
        """
        :param name:
            The name of the metric.

        :param scikit_kwargs:
            Custom arguments to be passed to scikit-learn 'mean_squared_error' function.
        """
        from sklearn.metrics import mean_squared_error
        super(MSE, self).__init__(metric_function=mean_squared_error, name=name, **scikit_kwargs)


class MAE(RegressionMetric):
    """Wrapper for scikit-learn 'mean_absolute_error' function."""

    def __init__(self, name: str = 'mae', **scikit_kwargs):
        """
        :param name:
            The name of the metric.

        :param scikit_kwargs:
            Custom arguments to be passed to scikit-learn 'mean_absolute_error' function.
        """
        from sklearn.metrics import mean_absolute_error
        super(MAE, self).__init__(metric_function=mean_absolute_error, name=name, **scikit_kwargs)


class R2(RegressionMetric):
    """Wrapper for scikit-learn 'r2_score' function."""

    def __init__(self, name: str = 'r2', **scikit_kwargs):
        """
        :param name:
            The name of the metric.

        :param scikit_kwargs:
            Custom arguments to be passed to scikit-learn 'r2_score' function.
        """
        from sklearn.metrics import r2_score
        super(R2, self).__init__(metric_function=r2_score, name=name, **scikit_kwargs)



"""Basic Backend Interface."""
import logging
from typing import Any, Union, List, Optional, Callable, Set

import numpy as np

from moving_targets.util.errors import not_implemented_message, BackendError


class Backend:
    """Basic Interface for a Moving Targets Master Backend."""

    _LOGGER: logging.Logger = logging.getLogger('Backend')
    """The inner backend logger instance."""

    _INDICATORS: Set[str] = {'>', '>=', '<', '<=', '==', '!='}

    @staticmethod
    def _transposed_axes(dim: int, index: int, place: int) -> List[int]:
        """Computes the order of the transposed axes.

        :param dim:
            The total number of dimensions.

        :param index:
            The index of the axis to move.

        :param place:
            The position where to move the given index. E.g., if dim = 4, index = 2, and place = 0, the third dimension
            will be brought to front thus it will be returned the list [2, 0, 1, 3]; otherwise, if dim = 4, index = 0,
            and place = 2, the operation will be inverted, thus returning [1, 2, 0, 3].

        :return:
            The list of indices of the transposed axes.
        """
        axes = list(range(dim))
        axes.insert(place, axes.pop(index))
        return axes

    @classmethod
    def _nested_names(cls, *keys: int, name: Optional[str]) -> List:
        """Generates an keys-dimensional list of nested lists containing the correct variables names when an array of
        variables is wanted instead of a single one by appending the variable position in the keys-dimensional tensor.

        :param keys:
            The number of keys per dimension of the array of variables.

        :param name:
            The common name for the variables.

        :return:
            The keys-dimensional structure of nested lists.
        """
        if len(keys) == 1:
            return [None if name is None else f'{name}_{i}' for i in range(keys[0])]
        else:
            return [cls._nested_names(*keys[1:], name=None if name is None else f'{name}_{i}') for i in range(keys[0])]

    @classmethod
    def _aux_warning(cls, exp: Optional[str], aux: Optional[str], msg: str):
        """Logs a warning on the Backend Logger when the 'aux' parameter is not used correctly.

        :param exp:
            The expected value of the aux parameter.

        :param aux:
            The actual value of the aux parameter.

        :param msg:
            The warning explanation.
        """
        if aux != 'auto' and exp != aux:
            cls._LOGGER.warning(f"'aux={aux}' has no effect since the solver needs {exp} auxiliary variables to {msg}.")

    def __init__(self, sum_fn: Callable = lambda v: np.sum(v)):
        """
        :param sum_fn:
            The backend function to perform a sum over a one-dimensional array of model variables.
        """
        super(Backend, self).__init__()

        self._sum_fn: Callable = sum_fn
        """The backend function to perform a sum over an array of model variables."""

        self.model: Optional = None
        """The inner model instance."""

        self.solution: Optional = None
        """The inner solution instance."""

    def _build_model(self) -> Any:
        """Initializes a model instance.

        :return:
            The model instance.
        """
        raise NotImplementedError(not_implemented_message(name='_build_model'))

    def _solve_model(self) -> Optional:
        """Solves the optimization problem.

        :return:
            The solution instance or None if no solution is found.
        """
        raise NotImplementedError(not_implemented_message(name='_solve_model'))

    def build(self) -> Any:
        """Initializes a model instance.

        :return:
            The backend itself.
        """
        if self.model is not None:
            self._LOGGER.warning("A model instance was already present and has been overwritten. If that was " +
                                 "intentional consider calling 'backend.clear()' before creating a new one.")
        self.model = self._build_model()
        self.solution = None
        return self

    def solve(self) -> Any:
        """Solves the optimization problem.

        :return:
            The backend itself.
        """
        self.solution = self._solve_model()
        return self

    def clear(self) -> Any:
        """Clears the backend resources.

        :return:
            The backend itself.
        """
        self.model = None
        self.solution = None
        return self

    def minimize(self, cost) -> Any:
        """Sets a cost function to minimize before solving.

        :param cost:
            The cost function.

        :return:
            The backend itself.
        """
        raise NotImplementedError(not_implemented_message(name='minimize'))

    def maximize(self, cost) -> Any:
        """Sets a cost function to maximize before solving.

        :param cost:
            The cost function.

        :return:
            The backend itself.
        """
        raise NotImplementedError(not_implemented_message(name='maximize'))

    def get_objective(self) -> float:
        """Gets the objective value of the solved model.

        :return:
            The objective value.
        """
        raise NotImplementedError(not_implemented_message(name='get_objective'))

    def get_values(self, expressions: np.ndarray) -> np.ndarray:
        """Gets the values of an array of expressions as found in the model solution.

        :param expressions:
            The array of expressions.

        :return:
            The array of solution values.
        """
        raise NotImplementedError(not_implemented_message(name='get_values'))

    def get_value(self, expression: Any) -> float:
        """Gets the value of an expressions as found in the model solution.

        :param expression:
            The expressions.

        :return:
            The solution value.
        """
        return self.get_values(np.array([expression]))[0]

    def add_binary_variables(self, *keys: int, name: Optional[str] = None) -> np.ndarray:
        """Creates an array of binary model variables.

        :param keys:
            The number of keys per dimension of the array of variables.

        :param name:
            The variables name.

        :return:
            The array of binary model variables.
        """
        return self.add_variables(*keys, vtype='binary', name=name, lb=0, ub=1)

    def add_binary_variable(self, name: Optional[str] = None) -> Any:
        """Creates a binary model variable.

        :param name:
            The variable name.

        :return:
            The binary model variable.
        """
        return self.add_variable(vtype='binary', name=name, lb=0, ub=1)

    def add_integer_variables(self,
                              *keys: int,
                              lb: float = 0,
                              ub: float = float('inf'),
                              name: Optional[str] = None) -> np.ndarray:
        """Creates an array of integer model variables.

        :param keys:
            The number of keys per dimension of the array of variables.

        :param lb:
            The variables lower bound.

        :param ub:
            The variables upper bound.

        :param name:
            The variables name.

        :return:
            The array of integer model variables.
        """
        return self.add_variables(*keys, vtype='integer', name=name, lb=lb, ub=ub)

    def add_integer_variable(self,
                             lb: float = 0,
                             ub: float = float('inf'),
                             name: Optional[str] = None) -> np.ndarray:
        """Creates an integer model variable.

        :param lb:
            The variable lower bound.

        :param ub:
            The variable upper bound.

        :param name:
            The variable name.

        :return:
            The integer model variable.
        """
        return self.add_variable(vtype='integer', name=name, lb=lb, ub=ub)

    def add_continuous_variables(self,
                                 *keys: int,
                                 lb: float = -float('inf'),
                                 ub: float = float('inf'),
                                 name: Optional[str] = None) -> np.ndarray:
        """Creates an array of continuous model variables.

        :param keys:
            The number of keys per dimension of the array of variables.

        :param lb:
            The variables lower bound.

        :param ub:
            The variables upper bound.

        :param name:
            The variables name.

        :return:
            The array of continuous model variables.
        """
        return self.add_variables(*keys, vtype='continuous', name=name, lb=lb, ub=ub)

    def add_continuous_variable(self,
                                lb: float = -float('inf'),
                                ub: float = float('inf'),
                                name: Optional[str] = None) -> np.ndarray:
        """Creates a continuous model variable.

        :param lb:
            The variable lower bound.

        :param ub:
            The variable upper bound.

        :param name:
            The variable name.

        :return:
            The continuous model variable.
        """
        return self.add_variable(vtype='continuous', name=name, lb=lb, ub=ub)

    def add_constant(self, val: Any, vtype: str = 'continuous', name: Optional[str] = None) -> Any:
        """Creates a model variable with a constant value.

        :param val:
            The numerical value of the constant.

        :param vtype:
            The constant type, by default this is 'continuous'.

        :param name:
            The constant name.

        :return:
            A model variable with value fixed to the given one via a constraint.
        """
        return self.add_variable(vtype=vtype, lb=val, ub=val, name=name)

    def add_constants(self, val: np.ndarray, vtype: str = 'continuous', name: Optional[str] = None) -> np.ndarray:
        """Creates an array of model variables with constant values.

        :param val:
            The array of numerical values of the constants.

        :param vtype:
            The constants type, by default this is 'continuous'.

        :param name:
            The constants name.

        :return:
            An array of model variables with value fixed to the given ones via a set of constraints.
        """
        # a default strategy to create an array of constants is to compute their names leveraging the utility function,
        # creating a mono-dimensional list of variables with correct names and fixed lower/upper bounds to the given
        # values then, eventually, reshaping then into the correct shape
        names = np.array(self._nested_names(*val.shape, name=name)).flatten()
        var = [self.add_variable(vtype=vtype, lb=v, ub=v, name=n) for v, n in zip(val.flatten(), names)]
        return np.reshape(var, val.shape)

    def add_variable(self, vtype: str, lb: float, ub: float, name: Optional[str] = None) -> Any:
        """Creates a model variable.

        :param vtype:
            The variables type, usually 'binary' or 'continuous'.

        :param lb:
            The variable lower bound.

        :param ub:
            The variable upper bound.

        :param name:
            The variable name.

        :return:
            The model variable.
        """
        raise NotImplementedError(not_implemented_message(name='add_variable'))

    def add_variables(self, *keys: int, vtype: str, lb: float, ub: float, name: Optional[str] = None) -> np.ndarray:
        """Creates an array of model variables.

        :param keys:
            The number of keys per dimension of the array of variables.

        :param vtype:
            The variables type, usually 'binary', 'integer', or 'continuous'.

        :param lb:
            The variables lower bound.

        :param ub:
            The variables upper bound.

        :param name:
            The variables name.

        :return:
            The array of model variables.
        """
        # a default strategy to create an array of variables is to compute their names leveraging the utility function
        # then creating a mono-dimensional list of variables and reshaping then into the correct shape
        names = np.array(self._nested_names(*keys, name=name)).flatten()
        return np.reshape([self.add_variable(vtype=vtype, lb=lb, ub=ub, name=n) for n in names], keys)

    def aux(self,
            expressions: Any,
            aux_vtype: Optional[str] = 'continuous',
            aux_lb: float = -float('inf'),
            aux_ub: float = float('inf'),
            aux_name: Optional[str] = None) -> Any:
        """If the 'aux_vtype' parameter is None, it simply return the input expressions, otherwise it builds and return
        an auxiliary variable (or an array of variables) having the given vtype and the respective other properties,
        which is equal to the given expression (or an array of expressions). Using auxiliary variables may come in
        handy when dealing with huge datasets since they can considerably speedup the model formulation; still,
        imposing equality constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :param expressions:
            Either a single expression or a array of expressions.

        :param aux_vtype:
            Either None (or 'auto') or the auxiliary variables type, usually 'binary', 'integer', or 'continuous'.

        :param aux_lb:
            The auxiliary variables lower bound.

        :param aux_ub:
            The auxiliary variables upper bound.

        :param aux_name:
            The auxiliary variables name.

        :return:
            Either a single auxiliary variable or the array of auxiliary variables.
        """
        if aux_vtype is None or aux_vtype == 'auto':
            return expressions
        else:
            if isinstance(expressions, np.ndarray):
                expressions, is_numpy = expressions, True
            else:
                expressions, is_numpy = np.array([expressions]), False
            variables = self.add_variables(*expressions.shape, vtype=aux_vtype, lb=aux_lb, ub=aux_ub, name=aux_name)
            self.add_constraints([v == e for v, e in zip(variables.flatten(), expressions.flatten())])
            return variables if is_numpy else variables[0]

    def add_constraints(self, constraints: Union[List, np.ndarray], name: Optional[str] = None) -> Any:
        """Adds a list of constraints to the model.

        :param constraints:
            The list of constraints.

        :param name:
            The constraints name.

        :return:
            The backend itself.
        """
        constraints = np.array(constraints)
        names = np.array(self._nested_names(*constraints.shape, name=name))
        for c, n in zip(constraints.flatten(), names.flatten()):
            self.add_constraint(constraint=c, name=n)
        return self

    def add_constraint(self, constraint, name: Optional[str] = None) -> Any:
        """Adds a constraint to the model.

        :param constraint:
            The constraint.

        :param name:
            The constraint name.

        :return:
            The backend itself.
        """
        raise NotImplementedError(not_implemented_message(name='add_constraint'))

    def add_indicator_constraints(self,
                                  indicators: np.ndarray,
                                  expressions: Union[List, np.ndarray],
                                  value: int = 1,
                                  name: Optional[str] = None) -> Any:
        """Impose a set of indicator constraints over the given expressions using the given binary indicators. The
        indicators indicators[i] are such that if indicators[i] == value (with value either 0 or 1), then the i-th
        expression holds, but in case indicators[i] == not value, then there is no information about that expression.

        :param indicators:
            An array of binary variables on which to impose the indicator constraints.

        :param expressions:
            An array or list of expressions for which it is necessary to check whether they hold or not.

        :param value:
            The value assumed by a binary variable when the respective expression holds, either 0 or 1.

        :param name:
            The constraints name.

        :return:
            The backend itself.

        :raise `BackendError`:
            If the backend cannot handle indicator variables.
        """
        expressions = np.array(expressions)
        names = np.array(self._nested_names(*expressions.shape, name=name))
        for i, e, n in zip(indicators.flatten(), expressions.flatten(), names.flatten()):
            self.add_indicator_constraint(indicator=i, expression=e, value=value, name=n)
        return self

    def add_indicator_constraint(self,
                                 indicator: Any,
                                 expression: Any,
                                 value: int = 1,
                                 name: Optional[str] = None) -> Any:
        """Impose an indicator constraint over the given expression using the given binary indicator. The indicator
        is such that if indicator == value (with value either 0 or 1), then the expression holds, but in case
        indicator == not value, then there is no information about the expression.

        :param indicator:
            The binary variable on which to impose the indicator constraints.

        :param expression:
            The expression for which it is necessary to check whether they hold or not.

        :param value:
            The value assumed by a binary variable when the respective expression holds, either 0 or 1.

        :param name:
            The constraint name.

        :return:
            The backend itself.

        :raise `BackendError`:
            If the backend cannot handle indicator variables.
        """
        raise BackendError(unsupported='indicator constraints')

    def is_greater(self, a: np.ndarray, b: Union[float, np.ndarray]) -> np.ndarray:
        """Builds auxiliary binary indicator variables which take value one if the expressions in the first array are
         greater than the expressions in the second array. Please note that this is enforced via indicator constraints
         so that if z[i] == 1 -> a[i] >= b[i] and if z[i] == 0 -> a[i] <= b[i], thus in case a[i] is strictly equal to
         b[i] the indicator variable z[i] can assume both values.

        :param a:
            The first array representing the left-hand sides.

        :param b:
            Either a single reference value or a second array of expressions representing the right-hand sides.

        :return:
            The array of binary indicator variables.

        :raise `BackendError`:
            If the backend cannot handle indicator variables.
        """
        # if b is an array of expressions, the i-th right hand side will be b[i] (with b flattened), otherwise it is b
        if isinstance(b, np.ndarray):
            b = b.flatten()
            rhs = lambda i: b[i]
        else:
            rhs = lambda i: b
        z = self.add_binary_variables(a.size)
        self.add_indicator_constraints(z, expressions=[lhs >= rhs(i) for i, lhs in enumerate(a.flatten())], value=1)
        self.add_indicator_constraints(z, expressions=[lhs <= rhs(i) for i, lhs in enumerate(a.flatten())], value=0)
        return z.reshape(a.shape)

    def is_less(self, a: np.ndarray, b: np.ndarray) -> np.ndarray:
        """Builds auxiliary binary indicator variables which take value one if the expressions in the first array are
         lower than the expressions in the second array. Please note that this is enforced via indicator constraints
         so that if z[i] == 1 -> a[i] <= b[i] and if z[i] == 0 -> a[i] >= b[i], thus in case a[i] is strictly equal to
         b[i] the indicator variable z[i] can assume both values.

        :param a:
            The first array.

        :param b:
            The second array.

        :return:
            The array of binary indicator variables.

        :raise `BackendError`:
            If the backend cannot handle indicator variables.
        """
        # if b is an array of expressions, the i-th right hand side will be b[i] (with b flattened), otherwise it is b
        if isinstance(b, np.ndarray):
            b = b.flatten()
            rhs = lambda i: b[i]
        else:
            rhs = lambda i: b
        z = self.add_binary_variables(a.size)
        self.add_indicator_constraints(z, expressions=[lhs <= rhs(i) for i, lhs in enumerate(a.flatten())], value=1)
        self.add_indicator_constraints(z, expressions=[lhs >= rhs(i) for i, lhs in enumerate(a.flatten())], value=0)
        return z.reshape(a.shape)

    def sum(self, a: np.ndarray, axis: Optional[int] = None, asarray: bool = False, aux: Optional[str] = 'auto') -> Any:
        """Computes the sum of an array of variables.

        :param a:
            An array of model variables.

        :param axis:
            The dimension on which to aggregate or None to aggregate the whole data.

        :param asarray:
            In case the aggregation returns a single expression, whether to return it as a zero-dimensional numpy array
            or as the expression itself.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The variables sum.
        """
        # if no axis is specified, simply leverage the internal sum function to get a single expression, otherwise
        # compute the sum over the requested axis and then reshape the output array accordingly
        if axis is None:
            expressions = self._sum_fn(a)
            expressions = np.reshape(expressions, ()) if asarray else expressions
        else:
            # check that the axis is in bound and handle negative axis values
            assert axis in range(-a.ndim, a.ndim), f"Axis {axis} is out of bound for array with {a.ndim} dimensions"
            axis = axis % a.ndim
            # transpose the array in order to bring the axis-th dimension to the back, then reshape it into a matrix so
            # that we can aggregate only on the last dimension, which is the one representing the chosen axis
            axes = self._transposed_axes(dim=a.ndim, index=axis, place=a.ndim - 1)
            expressions = a.transpose(axes).reshape((-1, a.shape[axis]))
            expressions = [np.sum(row) for row in expressions]
            # at this point, the "expressions" list will have size a.size / a.shape[axis], thus we need to reshape it
            # accordingly to the input shape by popping out the axis-th dimension, which is now one due to the sum
            # (also, take care that the output is at least one-dimensional, otherwise return a single expression)
            new_shape = list(a.shape)
            new_shape.pop(axis)
            expressions = np.reshape(expressions, new_shape) if len(new_shape) > 0 or asarray else expressions[0]
        return self.aux(expressions, aux_vtype=aux)

    def square(self, a: np.ndarray, aux: Optional[str] = 'auto') -> np.ndarray:
        """Computes the squared values over an array of variables.

        :param a:
            An array of model variables.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The array of squared values.

        :raise `BackendError`:
            If the backend cannot handle squared values.
        """
        raise BackendError(unsupported='squared values')

    def sqrt(self, a: np.ndarray, aux: Optional[str] = 'auto') -> np.ndarray:
        """Computes the squared roots over an array of variables.

        :param a:
            An array of model variables.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The array of squared roots.
        """
        raise BackendError(unsupported='squared roots')

    def abs(self, a: np.ndarray, aux: Optional[str] = 'auto') -> np.ndarray:
        """Computes the absolute values over an array of variables.

        :param a:
            An array of model variables.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The array of absolute values.

        :raise `BackendError`:
            If the backend cannot handle absolute values.
        """
        raise BackendError(unsupported='absolute values')

    def log(self, a: np.ndarray, aux: Optional[str] = 'auto') -> np.ndarray:
        """Computes the logarithms over an array of variables.

        :param a:
            An array of model variables.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The array of logarithms.

        :raise `BackendError`:
            If the backend cannot handle logarithms.
        """
        raise BackendError(unsupported='logarithms')

    def mean(self,
             a: np.ndarray,
             axis: Optional[int] = None,
             asarray: bool = False,
             aux: Optional[str] = 'auto') -> Any:
        """Computes the mean of an array of variables.

        :param a:
            An array of model variables.

        :param axis:
            The dimension on which to aggregate or None to aggregate the whole data.

        :param asarray:
            In case the aggregation returns a single expression, whether to return it as a zero-dimensional numpy array
            or as the expression itself.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The variables mean.
        """
        sum_expression = self.sum(a, axis=axis, asarray=asarray, aux=aux)
        num_aggregated = a.size if axis is None else a.shape[axis]
        return self.aux(sum_expression / num_aggregated, aux_vtype=aux)

    def var(self, a: np.ndarray, axis: Optional[int] = None, asarray: bool = False, aux: Optional[str] = 'auto') -> Any:
        """Computes the variance of an array of variables.

        :param a:
            An array of model variables.

        :param axis:
            The dimension on which to aggregate or None to aggregate the whole data.

        :param asarray:
            In case the aggregation returns a single expression, whether to return it as a zero-dimensional numpy array
            or as the expression itself.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The variables variance.
        """
        # if no axis is specified, simply compute the mean (which will be a single expression), then compute the array
        # of differences and shape it correctly, otherwise compute the means over the requested axis (which will be in
        # the form of an array), then compute the differences and eventually reshape them accordingly
        aux_mean, aux_squared = ('continuous', None) if aux == 'auto' else (aux, aux)
        mean_expressions = self.mean(a, axis=axis, asarray=True, aux=aux_mean)
        if axis is None:
            # we compute the array of differences element by element because of some problems that certain backends
            # arise due to the impossibility to subtract a single variable (or expression) to an array of variables
            diff_expressions = np.reshape([self.subtract(ai, mean_expressions) for ai in a.flatten()], a.shape)
        else:
            # check that the axis is in bound and handle negative axis values
            assert axis in range(-a.ndim, a.ndim), f"Axis {axis} is out of bound for array with {a.ndim} dimensions"
            axis = axis % a.ndim
            # transpose the array in order to bring the axis-th dimension to the front so to compute the differences of
            # each row in the given axis with respect to the mean of that axis, then transpose the axis from the front
            # back to its place since we want to match exactly the original shape
            a = a.transpose(self._transposed_axes(dim=a.ndim, index=axis, place=0))
            diff_expressions = np.array([self.subtract(ai, mean_expressions) for ai in a])
            diff_expressions = diff_expressions.transpose(self._transposed_axes(dim=a.ndim, index=0, place=axis))
        # we can finally build auxiliary variables for the differences (if needed), then square these differences and,
        # eventually, compute the mean of the squared differences over the given axis
        diff_expressions = self.aux(expressions=diff_expressions, aux_vtype=aux)
        squared_expressions = self.square(diff_expressions, aux=aux_squared)
        return self.mean(squared_expressions, axis=axis, asarray=asarray, aux=aux)

    def add(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = 'auto'):
        """Performs the pairwise sum between two arrays.

        :param a:
            The first array.

        :param b:
            The second array.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The array of pairwise sums between a and b.
        """
        return self.aux(expressions=a + b, aux_vtype=aux)

    def subtract(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = 'auto'):
        """Performs the pairwise subtraction between two arrays.

        :param a:
            The first array.

        :param b:
            The second array.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The array of pairwise differences between a and b.
        """
        return self.aux(expressions=a - b, aux_vtype=aux)

    def multiply(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = 'auto'):
        """Performs the pairwise product between two arrays.

        :param a:
            The first array.

        :param b:
            The second array.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The array of pairwise products between a and b.
        """
        return self.aux(expressions=a * b, aux_vtype=aux)

    def divide(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = 'auto'):
        """Performs the pairwise division between two arrays.

        :param a:
            The first array.

        :param b:
            The second array.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The array of pairwise divisions between a and b.
        """
        return self.aux(expressions=a / b, aux_vtype=aux)

    def dot(self,
            a: np.ndarray,
            b: np.ndarray,
            asarray: bool = False,
            aux: Optional[str] = 'auto') -> Union[Any, np.ndarray]:
        """Performs the dot product between two arrays.

        :param a:
            The first array (at most 2d).

        :param b:
            The second array (at most 2d).

        :param asarray:
            In case the aggregation returns a single expression, whether to return it as a zero-dimensional numpy array
            or as the expression itself.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The dot product between a and b.
        """
        dot_expression = self.aux(expressions=a @ b, aux_vtype=aux)
        return np.array(dot_expression) if asarray else dot_expression



"""Interfaces and classes for Moving Targets Masters."""

from moving_targets.masters import losses, optimizers, backends
from moving_targets.masters.masters import Master, RegressionMaster, ClassificationMaster



"""Basic Learner Interface."""
import time
from typing import Any, Optional, Set, Union, List

import numpy as np

from moving_targets.callbacks import StatsLogger
from moving_targets.util.errors import not_implemented_message
from moving_targets.util.scalers import Scaler
from moving_targets.util.typing import Dataset


class Learner(StatsLogger):
    """Basic interface for a Moving Targets Learner."""

    @staticmethod
    def _parameters() -> Set[str]:
        return {'elapsed_time'}

    def __init__(self,
                 x_scaler: Union[None, Scaler, str],
                 y_scaler: Union[None, Scaler, str],
                 stats: Union[bool, List[str]]):
        """
        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.
        """
        super(Learner, self).__init__(stats=stats, name='Learner')

        self.x_scaler: Optional[Scaler] = Scaler(default_method=x_scaler) if isinstance(x_scaler, str) else x_scaler
        """The (optional) scaler for the input data."""

        self.y_scaler: Optional[Scaler] = Scaler(default_method=y_scaler) if isinstance(y_scaler, str) else y_scaler
        """The (optional) scaler for the output data."""

        self._macs: Optional = None
        """Reference to the MACS object encapsulating the `Learner`."""

        self._time: Optional[float] = None
        """An auxiliary variable to keep track of the elapsed time between iterations."""

    def log(self, **cache):
        self._macs.log(**cache)

    def on_training_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self._macs = macs
        self._time = time.time()

    def on_training_end(self, macs, x, y: np.ndarray, p: np.ndarray, val_data: Optional[Dataset]):
        self._log_stats(elapsed_time=time.time() - self._time)
        self._time = None
        self._macs = None

    def fit(self, x, y: np.ndarray, sample_weight: Optional[np.ndarray] = None) -> Any:
        """Scales the (x, y) data and use it to fit the `Learner`.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param sample_weight:
            The (optional) array of sample weights.

        :return:
            The `Learner` itself.
        """
        x = x if self.x_scaler is None else self.x_scaler.fit_transform(data=x)
        y = y if self.y_scaler is None else self.y_scaler.fit_transform(data=y)
        self._fit(x=x, y=y, sample_weight=sample_weight)
        return self

    def predict(self, x) -> np.ndarray:
        """Uses the fitted `Learner` configuration to predict labels from input samples.

        :param x:
            The matrix/dataframe of input samples.

        :return:
            The vector of predicted labels.
        """
        x = x if self.x_scaler is None else self.x_scaler.transform(data=x)
        p = self._predict(x)
        return p if self.y_scaler is None else self.y_scaler.inverse_transform(data=p)

    def _fit(self, x, y: np.ndarray, sample_weight: Optional[np.ndarray] = None):
        """Implements the fitting strategy based on the kind of learner.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param sample_weight:
            The (optional) array of sample weights.
        """
        raise NotImplementedError(not_implemented_message(name='_fit'))

    def _predict(self, x) -> np.ndarray:
        """Implements the predictive strategy based on the kind of learner.

        :param x:
            The matrix/dataframe of input samples.

        :return:
            The vector of predicted labels.
        """
        raise NotImplementedError(not_implemented_message(name='_predict'))



"""History Callback"""
import fnmatch
import warnings
from typing import List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from moving_targets.callbacks.logger import Logger
from moving_targets.util.typing import Dataset


class History(Logger):
    """Collects the training information and eventually plots it."""

    def __init__(self):
        """"""
        super(History, self).__init__()

        self._history: List = []
        """An internal list which will be used to create the final DataFrame object at the end of the training."""

    def on_iteration_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        # Creates a dataframe from the cached items and appends them to a list indexed by iteration.
        self._history.append(pd.DataFrame([self._cache.values()], columns=self._cache.keys()))
        self._cache = {}

    def on_process_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        # Creates a single dataframe by concatenating the sub-dataframes from each iteration.
        if len(self._history) > 0:
            self._history = pd.concat(self._history, ignore_index=True)

    def plot(self,
             features: Union[None, int, str, List[str], List[List[Optional[str]]]] = '/',
             orient_rows: bool = False,
             tight_layout=True,
             figsize=(16, 9),
             **plt_kwargs):
        """Plots the training information which were previously collected in a dataframe.

        :param features:
            Indicates which feature to plot as well as the plotting strategy.
            If None is passed, all the (numeric) features are automatically displayed according to the figure ratio.
            If an integer is passed, all the (numeric) features are plotted and the parameter indicates number of row
            or columns in the final subplot, respectively to the value of the orient_rows argument.
            If a string is passed, this is used to display the subplots according to a common prefix (for instance, if
            features = '/' and the features are ['trn/loss', 'trn/metric', 'val/loss', 'val/metric'], then there will
            be a row or column for the 'trn' prefix and another one for the 'val' prefix).
            If a list of strings is passed, these strings are considered as patters, therefore there will be a row or
            column for each feature matching with the pattern (each pattern can be either the exact name of a feature
            or it can contain a wildcard, e.g., '*/mse' or 'predictions/*').
            If a list of lists is passed, it displays the subplots accordingly to each entry in the grid (if an entry
            is None, the respective subplot will be left empty).

        :param orient_rows:
            Whether to orient the plots by column or by row. This influences also the 'num_subplots' parameter since
            when 'orient_rows' is set to True the 'num_subplots' indicates the number of rows while when 'orient_rows'
            is set to False the 'num_subplots' indicates the number of columns.

        :param tight_layout:
            Matplotlib `figure()` argument.

        :param figsize:
            Matplotlib `figure()` argument.

        :param plt_kwargs:
            Additional plot arguments to be passed to `figure()`.
        """
        if not isinstance(self._history, pd.DataFrame):
            warnings.warn('Process did not end correctly, therefore no dataframe can be plotted')
            return
        # HANDLE COLUMNS TO BE PLOTTED (FEATURES)
        # 1. first of all, retrieve all the columns that can be plotted (i.e., they are numeric)
        # 2. if <features> is None, compute the best subplots disposition according to their number, the figure ratio,
        #    and the orientation, so that it is brought back to the case in which <features> is an integer
        # 3. if <features> is a string, retrieve all the prefixes matching that separator and store them into a list,
        #    so that it is brought back to the case in which <features> is a list of patterns
        columns = [c for c in self._history.columns if np.issubdtype(self._history[c].dtype, np.number)]
        if features is None:
            ratio = figsize[0] / figsize[1] if orient_rows else figsize[1] / figsize[0]
            features = int(max(np.sqrt(ratio * len(columns)).round(), 1))
        elif isinstance(features, str):
            prefixes = {f'{column.split(features)[0]}' for column in columns}
            features = [f'{prefix}{features}*' for prefix in prefixes]
        # HANDLE THE SUBPLOT POSITIONS
        # 1. if <features> is an integer, it arranges the subplots by taking them in groups of <features>; moreover the
        #    orientation behaviour is inverted since columns are taken by <features> steps
        # 2. otherwise, <features> must be either a list of strings or a list of lists; in the first case, each string
        #    is mapped to the list of columns matching for the pattern, otherwise it means that this is the list of
        #    lists directly passed by the user, thus some integrity checks are performed
        # 3. eventually, the list of lists is converted into a well-shaped numpy array by filling up each row to the
        #    length of the maximal rows since some of them may have less subplots than the maximum one
        if isinstance(features, int):
            columns = [columns[i::features] for i in range(features)]
            orient_rows = not orient_rows
        else:
            for i, entry in enumerate(features):
                if isinstance(entry, str):
                    features[i] = fnmatch.filter(columns, entry)
                else:
                    for f in entry:
                        if f is not None:
                            assert f in columns, f"Feature '{f} is either not present in history or it is not numeric"
            columns = features
        num_rows = len(columns)
        num_cols = np.max([len(row) for row in columns])
        positions = np.array([[None] * num_cols] * num_rows)
        for i, row in enumerate(columns):
            positions[i, :len(row)] = row
        positions = positions if orient_rows else positions.transpose()
        # PLOT EACH COLUMN IN A SUBPLOT
        #   1. iterates by rows and columns
        #   2. if the column is not None, it plots the data
        ax = None
        num_rows, num_cols = positions.shape
        plt.figure(figsize=figsize, tight_layout=tight_layout, **plt_kwargs)
        for idx, column in enumerate(positions.flatten()):
            if column is not None:
                ax = plt.subplot(num_rows, num_cols, idx + 1, sharex=ax)
                x, y = self._history.index, self._history[column]
                plt.plot(x, y)
                plt.scatter(x, y)
                ax.set(title=column, xlabel='', ylabel='')
                # fix integer x_ticks based on autogenerated ones in order to avoid superimposed values
                ticks = np.unique(ax.get_xticks().round().astype(int))
                ax.set_xticks([t for t in ticks if t in range(x.min(), x.max() + 1)])
        plt.show()



# Run this script from the origin folder as:
#   > "python setup.py clean" in order to clean previous builds
#   > "python setup.py test" in order to execute all the unittests
#   > "python setup.py sdist" in order to build the library
#
# The package can then be published with:
#   > twin upload dist/*

from setuptools import find_packages, setup

# set up the library metadata and make the build
with open('README.md', 'r') as readme:
    setup(
        name='moving-targets',
        version='0.2.1',
        maintainer='Luca Giuliani',
        maintainer_email='luca.giuliani13@unibo.it',
        author='University of Bologna - DISI',
        description='Moving Targets: a framework for constrained machine learning',
        long_description=readme.read(),
        long_description_content_type='text/markdown',
        packages=find_packages(include=['moving_targets*']),
        python_requires='>=3.7',
        install_requires=['matplotlib~=3.4.3', 'numpy~=1.21.4', 'pandas~=1.3.4', 'scikit-learn~=1.0.1'],
        test_suite='test'
    )
